import errno
from io import BytesIO
import logging
import os
import pytest
import stat
import struct
import time
import threading
from unittest.mock import patch, MagicMock
from uuid import uuid4

import confluent_kafka

from adc.errors import KafkaException
from adc.kafka import parse_kafka_url

import hop
from hop.robust_publisher import _RAPriorityQueue, PublicationJournal, RobustProducer


logger = logging.getLogger("hop")


def test_queue_construction():
    q = _RAPriorityQueue()
    assert len(q) == 0  # queue should initially be empty


def test_queue_len():
    q = _RAPriorityQueue()
    expected = 0

    def add(key, value):
        nonlocal expected
        q.insert(key, value)
        expected += 1

    def remove():
        nonlocal expected
        result = q.pop_highest_priority()
        if result is not None:
            expected -= 1

    add(1, "a")
    assert len(q) == expected
    add(2, "b")
    assert len(q) == expected
    remove()
    assert len(q) == expected
    add(0, "c")
    assert len(q) == expected
    remove()
    assert len(q) == expected
    remove()
    assert len(q) == expected
    remove()
    assert len(q) == expected


def test_queue_contains():
    q = _RAPriorityQueue()
    assert "foo" not in q
    q.insert("foo", "bar")
    assert "foo" in q
    q.pop_highest_priority()
    assert "foo" not in q
    q.insert("baz", "quux")
    q.insert("xen", "hom")
    assert "baz" in q
    assert "xen" in q


def test_queue_insert_setitem():
    q = _RAPriorityQueue()
    assert "foo" not in q
    q.insert("foo", "bar")
    assert "foo" in q
    assert q["foo"] == "bar"
    q.insert("foo", "baz")
    assert q["foo"] == "baz"
    q["foo"] = "quux"
    assert q["foo"] == "quux"
    q["bar"] = "xen"
    assert q["bar"] == "xen"


def test_queue_getitem():
    q = _RAPriorityQueue()
    with pytest.raises(KeyError):
        q["foo"]
    q.insert("foo", "bar")
    assert q["foo"] == "bar"
    q.insert("baz", "quux")
    assert q["foo"] == "bar"
    assert q["baz"] == "quux"
    q.pop_highest_priority()
    assert q["foo"] == "bar"
    with pytest.raises(KeyError):
        q["baz"]


def test_queue_pop():
    q = _RAPriorityQueue()
    assert q.pop_highest_priority() is None
    # if there is only one item, pop must remove it
    q.insert(1, "foo")
    assert q.pop_highest_priority() == (1, "foo")
    # multiple items must be removed in priority order
    q.insert(2, "bar")
    q.insert(3, "baz")
    q.insert(1, "foo")
    assert q.pop_highest_priority() == (1, "foo")
    assert q.pop_highest_priority() == (2, "bar")
    assert q.pop_highest_priority() == (3, "baz")


def test_queue_remove_del():
    q = _RAPriorityQueue()
    with pytest.raises(KeyError):
        q.remove(1)
    q.insert(2, "bar")
    q.insert(3, "baz")
    q.insert(1, "foo")
    q.remove(2)
    assert 2 not in q
    assert q.pop_highest_priority() == (1, "foo")
    assert q.pop_highest_priority() == (3, "baz")
    with pytest.raises(KeyError):
        del q[1]
    q.insert(2, "bar")
    q.insert(3, "baz")
    q.insert(1, "foo")
    del q[1]
    assert 1 not in q
    assert q.pop_highest_priority() == (2, "bar")
    assert q.pop_highest_priority() == (3, "baz")


########################################


def test_journal_construct_no_file(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)
    assert not j.has_messages_to_send()
    assert not j.has_messages_in_flight()
    assert j.get_next_message_to_send() is None
    assert os.path.exists(journal_path)


def test_journal_construct_empty_file(tmpdir):
    journal_path = tmpdir.join("journal")
    open(journal_path, "ab")  # touch file
    j = PublicationJournal(journal_path)
    assert not j.has_messages_to_send()
    assert j.get_next_message_to_send() is None
    assert not j.has_messages_in_flight()
    assert os.path.exists(journal_path)


def test_journal_construct_bad_perms_file(tmpdir):
    journal_path = tmpdir.join("journal")
    open(journal_path, "ab")  # touch file
    os.chmod(journal_path, stat.S_IWUSR)  # remove read permission
    with pytest.raises(PermissionError):
        j = PublicationJournal(journal_path)


def test_journal_queue_message(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    j.queue_message(b"message 0", "topic1")
    assert j.has_messages_to_send()
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0
    s0 = j.get_next_message_to_send()
    assert s0[1] == b"message 0"
    assert s0[2] == "topic1"
    assert s0[3] is None
    assert s0[4] is None
    assert not j.has_messages_to_send()
    assert j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic1" in j.topics_with_messages_in_flight()

    headers = [(b"header_0", b"value_0"), (b"header_1", b"value_1")]
    j.queue_message(b"message 1", "topic2", headers=headers)
    assert j.has_messages_to_send()
    s1 = j.get_next_message_to_send()
    assert not j.has_messages_to_send()
    assert s1[1] == b"message 1"
    assert s1[2] == "topic2"
    assert s1[3] == headers
    assert s1[4] is None
    assert len(j.topics_with_messages_in_flight()) == 2
    assert "topic1" in j.topics_with_messages_in_flight()
    assert "topic2" in j.topics_with_messages_in_flight()

    j.queue_message(b"message 2", "topic2", key=b"some_key")
    assert j.has_messages_to_send()
    s2 = j.get_next_message_to_send()
    assert not j.has_messages_to_send()
    assert s2[1] == b"message 2"
    assert s2[2] == "topic2"
    assert s2[3] is None
    assert s2[4] == b"some_key"
    assert len(j.topics_with_messages_in_flight()) == 2
    assert "topic1" in j.topics_with_messages_in_flight()
    assert "topic2" in j.topics_with_messages_in_flight()


def test_journal_queue_message_write_failure(tmpdir):
    def write_fail(*args, **kwargs):
        raise OSError(errno.EIO, os.strerror(errno.EIO))
    file_mock = MagicMock()
    file_mock.write = write_fail
    file_mock.__enter__ = MagicMock(return_value=file_mock)
    open_mock = MagicMock(return_value=file_mock)

    journal_path = tmpdir.join("journal")
    with patch("builtins.open", open_mock):
        j = PublicationJournal(journal_path)
        with pytest.raises(RuntimeError) as excinfo:
            j.queue_message(b"some data", "topic")
        assert "Failed to append record to journal" in str(excinfo.value)


def test_journal_mark_sent(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    j.queue_message(b"message 0", "topic0")
    assert j.has_messages_to_send()
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0
    s0 = j.get_next_message_to_send()
    assert s0[1] == b"message 0"
    assert j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic0" in j.topics_with_messages_in_flight()

    j.mark_message_sent(s0[0])
    assert not j.has_messages_to_send()
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0

    with pytest.raises(RuntimeError):
        j.mark_message_sent(s0[0])  # can't mark the same message sent twice

    j.queue_message(b"message 1", "topic1")
    j.queue_message(b"message 2", "topic1")
    assert j.has_messages_to_send()
    assert len(j.topics_with_messages_in_flight()) == 0

    with pytest.raises(RuntimeError):
        j.mark_message_sent(1)  # can't mark things sent that haven't been extracted

    s1 = j.get_next_message_to_send()
    s2 = j.get_next_message_to_send()
    assert not j.has_messages_to_send()
    assert j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic1" in j.topics_with_messages_in_flight()

    # can mark messages sent out of order
    j.mark_message_sent(s2[0])
    assert j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic1" in j.topics_with_messages_in_flight()
    j.mark_message_sent(s1[0])
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0


def test_journal_mark_sent_write_failure(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)
    j.queue_message(b"message 0", "topic")
    del j

    def write_fail(*args, **kwargs):
        raise OSError(errno.EIO, os.strerror(errno.EIO))
    file_mock = MagicMock()
    file_mock.write = write_fail
    file_mock.__enter__ = MagicMock(return_value=file_mock)
    orig_open = open
    # allow reading to proceed normally so we can read the existing journal,
    # but intercept writing to inject failures

    def sneaky_open(path, mode, *args, **kwargs):
        if 'w' in mode or 'a' in mode:
            return file_mock
        else:
            return orig_open(path, mode, *args, **kwargs)

    with patch("builtins.open", wraps=sneaky_open):
        j = PublicationJournal(journal_path)
        s = j.get_next_message_to_send()
        with pytest.raises(RuntimeError) as excinfo:
            j.mark_message_sent(s[0])
        assert "Failed to append record to journal" in str(excinfo.value)


def test_journal_mark_sent_inflight_topics_desync(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)
    j.queue_message(b"message 0", "topic")
    s = j.get_next_message_to_send()
    assert len(j.inflight_topics) == 1
    assert "topic" in j.inflight_topics
    # Here we're trying to test that mark_message_sent detects when the expected entry in
    # inflight_topics. This is never supposed to happen, and various tests try to check that it
    # doesn't, but if there is a logic error missed by all other tests we want to get a readable
    # error message about this. There is, however, no obvious way to trigger this (since it is
    # supposed to be impossobile). Do it here by directly messing up the data structure.
    del j.inflight_topics["topic"]
    with pytest.raises(Exception) as excinfo:
        j.mark_message_sent(s[0])
    assert "Message to topic topic completed sending" in str(excinfo.value)
    assert "no messages were known to be in flight to that topic" in str(excinfo.value)


def test_journal_requeue_message(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    with pytest.raises(RuntimeError):
        j.requeue_message(0)  # can't requeue unknown messages

    j.queue_message(b"message 0", "topic")
    with pytest.raises(RuntimeError):
        j.requeue_message(0)  # can't requeue unsent messages

    assert len(j.topics_with_messages_in_flight()) == 0
    s0 = j.get_next_message_to_send()
    assert not j.has_messages_to_send()
    assert j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic" in j.topics_with_messages_in_flight()
    j.requeue_message(s0[0])
    assert j.has_messages_to_send()
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0

    with pytest.raises(RuntimeError):
        j.requeue_message(s0[0])  # can't requeue the same message twice at a time

    s0_2 = j.get_next_message_to_send()
    assert s0_2 == s0, "Message to be resent should match original"
    j.mark_message_sent(s0_2[0])

    with pytest.raises(RuntimeError):
        j.requeue_message(s0[0])  # can't requeue a sent message


def test_journal_ecoder_failure():
    format = "!I"

    def even_number_decoder(buffer):
        if buffer[-1] & 1:
            raise RuntimeError("Data does not represent an even number")
        return struct.unpack(format, buffer)[0]

    good_data = struct.pack(format, 222)
    bad_data = struct.pack(format, 32767)

    PublicationJournal._decode_raw_data(good_data, even_number_decoder, 0, "my value")
    with pytest.raises(RuntimeError) as err:
        PublicationJournal._decode_raw_data(bad_data, even_number_decoder, 0, "my value")
    assert "my value" in str(err.value)
    assert "Data does not represent an even number" in str(err.value)


def test_journal_restore_state(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    del j
    j = PublicationJournal(journal_path)  # reload empty journal
    assert not j.has_messages_to_send()
    assert not j.has_messages_in_flight()

    m0 = b"message 0"
    headers = [("header_0", b"value_0"), ("header_1", b"value_1")]
    j.queue_message(m0, "topic", headers, "mkey")
    del j
    j = PublicationJournal(journal_path)  # reload with message in queue
    assert j.has_messages_to_send()
    assert not j.has_messages_in_flight()

    s0 = j.get_next_message_to_send()
    assert s0[1] == m0, "Message body should be preserved by on-disk journal"
    assert s0[2] == "topic", "Message destination topic should be preserved by on-disk journal"
    assert s0[3] == headers, "Message headers should be preserved by on-disk journal"
    assert s0[4] == b"mkey", "Message key should be preserved by on-disk journal"
    del j
    j = PublicationJournal(journal_path)  # reload with message in flight
    # this may appear counter-intuitive, but if the journal/sender is offline while the message is
    # in flight, we have no way of knowing whether it arrived. Therefore, the conservative thing to
    # do is assume that it did not, and send it again.
    assert j.has_messages_to_send()
    assert not j.has_messages_in_flight()

    s0_2 = j.get_next_message_to_send()
    assert s0_2 == s0, "Message to be resent should match original"
    j.mark_message_sent(s0_2[0])
    del j
    j = PublicationJournal(journal_path)  # reload after message successfully sent
    assert not j.has_messages_to_send()
    assert not j.has_messages_in_flight()


def test_journal_restore_message_v1(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    sn = 0
    m0 = b"message 0"
    headers = [("header_0", b"value_0"), ("header_1", b"value_1")]

    # PublicationJournal no longer implements writing v1 mesage records,
    # so just do it manually here
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(sn))
    rbody.write(PublicationJournal.encode_int(len(m0)))  # message size
    rbody.write(m0)  # message data
    rbody.write(PublicationJournal.encode_int(len(headers)))  # number of headers
    for header in headers:
        # data must be encoded to be written
        hkey = header[0].encode("utf-8")
        value = header[1]
        rbody.write(PublicationJournal.encode_int(len(hkey)))  # header key length
        rbody.write(hkey)  # header key
        rbody.write(PublicationJournal.encode_int(len(value)))  # header value length
        rbody.write(value)  # header value
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    del j
    j = PublicationJournal(journal_path)  # reload with message in queue
    assert j.has_messages_to_send()
    assert not j.has_messages_in_flight()

    s0 = j.get_next_message_to_send()
    assert s0[1] == m0, "Message body should be preserved by on-disk journal"
    assert s0[2] is None, "V1 message records did not record a destination topic"
    assert s0[3] == headers, "Message headers should be preserved by on-disk journal"
    assert s0[4] is None, "V1 message records did not record a key"


def test_journal_restore_state_corrupted(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    headers = [("header_0", b"value_0"), ("header_1", b"value_1")]
    for i in range(0, 10):
        message = f"message {i}".encode("utf-8")
        if (i % 3) == 0:
            j.queue_message(message, f"topic{i % 2}", headers=headers)
        else:
            j.queue_message(message, f"topic{i % 2}")

    for i in range(0, 10):
        m = j.get_next_message_to_send()
        if (i % 2) == 0:
            j.mark_message_sent(m[0])

    del j
    # capture the raw journal for later manipulation
    with open(journal_path, "rb") as journal_file:
        journal_data = journal_file.read()

    # the unmodified journal should be readable
    j = PublicationJournal(journal_path)
    assert j.has_messages_to_send()
    assert len(j.messages_to_send) == 5
    assert not j.has_messages_in_flight()
    del j

    # test that corruption in any byte is detected
    for i in range(0, len(journal_data)):
        corrupted_data = bytearray(journal_data)
        corrupted_data[i] ^= 0b01010101  # flip some bits
        with open(journal_path, "wb") as journal_file:
            journal_file.write(corrupted_data)
        with pytest.raises(RuntimeError):
            try:
                j = PublicationJournal(journal_path)
            except RuntimeError as e:
                print(i, e)
                raise

    # test that invalid unicode in message header keys is detected
    corrupted_data = bytearray(journal_data).replace(b"header_0",
                                                     b"\xc3\x28\xa0\xa1\xf0\x28\x8c\xbc")
    with open(journal_path, "wb") as journal_file:
        journal_file.write(corrupted_data)
    with pytest.raises(RuntimeError):
        try:
            j = PublicationJournal(journal_path)
        except RuntimeError as e:
            print("invalid unicode message header", e)
            raise


def test_journal_restore_state_truncated(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    headers = [(b"header_0", b"value_0"), (b"header_1", b"value_1")]
    for i in range(0, 10):
        message = f"message {i}".encode("utf-8")
        if (i % 3) == 0:
            j.queue_message(message, f"topic{i % 2}", headers=headers)
        else:
            j.queue_message(message, f"topic{i % 2}")

    for i in range(0, 10):
        m = j.get_next_message_to_send()
        if (i % 2) == 0:
            j.mark_message_sent(m[0])

    del j
    # capture the raw journal for later manipulation
    with open(journal_path, "rb") as journal_file:
        journal_data = journal_file.read()

    # these are the offsets of the ends of the records
    # if the file format or test data is changed, these must be updated
    # truncating the file between records is valid/undetectable, so we should not test these
    valid_break_points = [141, 220, 299, 440, 519, 598, 739, 818, 897, 1038,
                          1070, 1102, 1134, 1166, 1198]
    # test that truncation at any byte (in a record) is detected
    for i in range(1, len(journal_data)):
        if i in valid_break_points:
            continue
        with open(journal_path, "wb") as journal_file:
            journal_file.write(journal_data[0:i])
        with pytest.raises(RuntimeError):
            try:
                j = PublicationJournal(journal_path)
            except RuntimeError as e:
                print(i, e)
                raise
            print(f"Failure: Truncation at byte {i} should be detected")


def test_journal_restore_duplicate_sequence_number_to_send(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"data"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(58))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    # record the same message again, duplicating the sequence number
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Duplicate message sequence number" in str(excinfo.value)

    journal_path = tmpdir.join("journal2")
    j = PublicationJournal(journal_path)
    test_message2 = b"other data"
    test_topic = b"a_topic"
    rbody2 = BytesIO()
    rbody2.write(PublicationJournal.encode_int(58))  # sequence number
    rbody2.write(PublicationJournal.encode_int(len(test_topic)))  # topic name length
    rbody2.write(test_topic)  # topic name
    rbody2.write(PublicationJournal.encode_int(len(test_message2)))  # data length
    rbody2.write(test_message2)  # data
    rbody2.write(PublicationJournal.encode_int(0))  # no key
    rbody2.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    # record another message but with the same sequence number
    j._write_record(PublicationJournal.msg_record_type_v2, rbody2.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Duplicate message sequence number" in str(excinfo.value)


def test_journal_restore_too_short_message_record_v1(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # a message record must contain the sequence number, mesage length, an header count
    # requiring 3*8 = 24 bytes. Anything shorter cannot be valid.
    j._write_record(PublicationJournal.msg_record_type_v1, b"tooshort")
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "too small to conain required data for record" in str(excinfo.value)


def test_journal_restore_too_short_message_record_v2(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # a message record must contain the sequence number, mesage length, an header count
    # requiring 3*8 = 24 bytes. Anything shorter cannot be valid.
    j._write_record(PublicationJournal.msg_record_type_v2, b"tooshort")
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "too small to conain required data for record" in str(excinfo.value)


def test_journal_restore_mismatched_body_length_v1(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"somedata"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    # use wrong data length
    rbody.write(PublicationJournal.encode_int(len(test_message) * 2))
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Claimed message data length" in str(excinfo.value)
    assert "exceeds record body length" in str(excinfo.value)


def test_journal_restore_missing_headers_v1(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"data"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    # claim a bunch of headers, but then don't include any
    rbody.write(PublicationJournal.encode_int(56))
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Claimed number of message headers" in str(excinfo.value)
    assert "exceeds remaining space in record body" in str(excinfo.value)


def test_journal_restore_mismatched_topic_length_v2(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"somedata"
    test_topic = b"a_topic"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    # use wrong topic name length
    rbody.write(PublicationJournal.encode_int(len(test_topic) * 4))
    rbody.write(test_topic)  # topic name
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no key
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v2, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Claimed topic name length" in str(excinfo.value)
    assert "exceeds record body length" in str(excinfo.value)


def test_journal_restore_invalid_topic_v2(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"somedata"
    invalid_utf8 = b"\x80\xC0\x0E\x00\xFF"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(invalid_utf8)))  # topic name length
    rbody.write(invalid_utf8)  # topic name
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no key
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v2, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Topic name is not valid UTF-8" in str(excinfo.value)


def test_journal_restore_mismatched_body_length_v2(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"somedata"
    test_topic = b"a_topic"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(test_topic)))  # topic name length
    rbody.write(test_topic)  # topic name
    # use wrong data length
    rbody.write(PublicationJournal.encode_int(len(test_message) * 2))
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no key
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v2, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Claimed message data length" in str(excinfo.value)
    assert "exceeds record body length" in str(excinfo.value)


def test_journal_restore_mismatched_key_length_v2(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"somedata"
    test_topic = b"a_topic"
    test_key = b"keykeykey"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(test_topic)))  # topic name length
    rbody.write(test_topic)  # topic name
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    # use wrong key legnth
    rbody.write(PublicationJournal.encode_int(len(test_key) * 2))
    rbody.write(test_key)  # key
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v2, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Claimed key length" in str(excinfo.value)
    assert "exceeds record body length" in str(excinfo.value)


def test_journal_restore_missing_headers_v2(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"data"
    test_topic = b"a_topic"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(0))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(test_topic)))  # topic name length
    rbody.write(test_topic)  # topic name
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no key
    # claim a bunch of headers, but then don't include any
    rbody.write(PublicationJournal.encode_int(56))
    j._write_record(PublicationJournal.msg_record_type_v2, rbody.getvalue())
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Claimed number of message headers" in str(excinfo.value)
    assert "exceeds remaining space in record body" in str(excinfo.value)


def test_journal_restore_sent_unknown_sequence_number(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    # record that a message was sent without it being previously mentioned
    j._write_record(PublicationJournal.sent_record_type, PublicationJournal.encode_int(56))
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Record of sent message" in str(excinfo.value)
    assert "which did not previously appear" in str(excinfo.value)


def test_journal_restore_duplicate_send(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # directly use internal recording function to generate records with valid checksums
    # but dubious meanings
    test_message = b"data"
    rbody = BytesIO()
    rbody.write(PublicationJournal.encode_int(17))  # sequence number
    rbody.write(PublicationJournal.encode_int(len(test_message)))  # data length
    rbody.write(test_message)  # data
    rbody.write(PublicationJournal.encode_int(0))  # no headers
    j._write_record(PublicationJournal.msg_record_type_v1, rbody.getvalue())
    # record that the message was sent
    j._write_record(PublicationJournal.sent_record_type, PublicationJournal.encode_int(17))
    # record that the message was sent _again_
    j._write_record(PublicationJournal.sent_record_type, PublicationJournal.encode_int(17))
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert "Record of sent message" in str(excinfo.value)
    assert "which did not previously appear" in str(excinfo.value)


def test_journal_restore_duplicate_bogus_record_type(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)
    # use a record type which is not defined
    bogus_record_type = 46000
    j._write_record(bogus_record_type, b"irrelevant")
    del j

    with pytest.raises(RuntimeError) as excinfo:
        j = PublicationJournal(journal_path)
    assert f"Invalid record type ({bogus_record_type})" in str(excinfo.value)


def test_journal_restore_read_failure(tmpdir):
    def read_fail(*args, **kwargs):
        raise OSError(errno.EIO, os.strerror(errno.EIO))
    file_mock = MagicMock()
    file_mock.read = read_fail
    file_mock.__enter__ = MagicMock(return_value=file_mock)
    open_mock = MagicMock(return_value=file_mock)

    journal_path = tmpdir.join("journal")
    with open(journal_path, 'w'):
        pass  # touch the file
    with patch("builtins.open", open_mock):
        with pytest.raises(RuntimeError) as excinfo:
            j = PublicationJournal(journal_path)
        assert "Journal corrupted: Unable to read" in str(excinfo.value)


def test_journal_get_delivery_callback_unqueued_message(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    # can't get a callback for a message which isn't yet queued
    with pytest.raises(RuntimeError) as excinfo:
        j.get_delivery_callback(22)
    assert "Cannot produce a delivery callback for message" in str(excinfo.value)
    assert "which is not in flight" in str(excinfo.value)

    # can't get a callback for a message which has already been sent
    j.queue_message(b"message 0", "topic")
    s0 = j.get_next_message_to_send()
    j.mark_message_sent(s0[0])

    with pytest.raises(RuntimeError) as excinfo:
        j.get_delivery_callback(s0[0])
    assert "Cannot produce a delivery callback for message" in str(excinfo.value)
    assert "which is not in flight" in str(excinfo.value)


def test_journal_delivery_callback_mark_sent(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    j.queue_message(b"message 0", "topic")
    s0 = j.get_next_message_to_send()
    callback = j.get_delivery_callback(s0[0])
    assert j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic" in j.topics_with_messages_in_flight()

    msg_obj = MagicMock()
    msg_obj.error = MagicMock(return_value=None)  # indicate no error if asked
    callback(None, msg_obj)  # no error and the 'message'
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0


def test_journal_delivery_callback_requeue(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    j.queue_message(b"message 0", "topic")
    s0 = j.get_next_message_to_send()
    callback = j.get_delivery_callback(s0[0])
    assert j.has_messages_in_flight()
    assert not j.has_messages_to_send()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic" in j.topics_with_messages_in_flight()

    msg_obj = MagicMock()
    err = confluent_kafka.KafkaError(confluent_kafka.KafkaError.NETWORK_EXCEPTION, "network error",
                                     False, True, False)
    msg_obj.error = MagicMock(return_value=err)

    # invoking the callback with an error should cause the message to be requeued
    callback(err, msg_obj)  # an error and the 'message'
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0
    assert j.has_messages_to_send()

    s0 = j.get_next_message_to_send()  # pretend we're retrying to send

    callback(None, msg_obj)  # no direct error, but the 'message' contains one
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0
    assert j.has_messages_to_send()


def test_journal_delivery_callback_nonretriable(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    j.queue_message(b"message 0", "topic")
    s0 = j.get_next_message_to_send()
    callback = j.get_delivery_callback(s0[0])
    assert j.has_messages_in_flight()
    assert not j.has_messages_to_send()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic" in j.topics_with_messages_in_flight()

    msg_obj = MagicMock()
    err = confluent_kafka.KafkaError(confluent_kafka.KafkaError.INVALID_REQUEST, "invalid request",
                                     False, False, False)
    msg_obj.error = MagicMock(return_value=err)

    # invoking the callback with an error should *not* cause the message to be requeued
    callback(err, msg_obj)  # an error and the 'message'
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0
    assert not j.has_messages_to_send()

    j.queue_message(b"message 0", "topic")
    s0 = j.get_next_message_to_send()
    callback = j.get_delivery_callback(s0[0])
    assert j.has_messages_in_flight()
    assert not j.has_messages_to_send()
    assert len(j.topics_with_messages_in_flight()) == 1
    assert "topic" in j.topics_with_messages_in_flight()

    callback(None, msg_obj)  # no direct error, but the 'message' contains one
    assert not j.has_messages_in_flight()
    assert len(j.topics_with_messages_in_flight()) == 0
    assert not j.has_messages_to_send()


class TrialLock():
    def __init__(self):
        self.held = False
        self.was_locked = False

    def __enter__(self):
        assert not self.held
        self.held = True
        self.was_locked = True

    def __exit__(self, exc_type, exc_value, exc_traceback):
        assert self.held
        self.held = False


def test_journal_delivery_callback_locking(tmpdir):
    journal_path = tmpdir.join("journal")
    j = PublicationJournal(journal_path)

    myLock = TrialLock()

    def wrap(func):
        def require_lock(*args, **kwargs):
            assert myLock.held, "Lock must be held for this operation"
            return func(*args, **kwargs)
        return require_lock

    j.queue_message(b"message 0", "topic")
    s0 = j.get_next_message_to_send()
    callback = j.get_delivery_callback(s0[0], myLock)

    msg_obj = MagicMock()
    msg_obj.error = MagicMock(return_value=None)
    err = confluent_kafka.KafkaError(confluent_kafka.KafkaError.NETWORK_EXCEPTION, "network error",
                                     False, True, False)

    with patch("hop.robust_publisher.PublicationJournal.mark_message_sent",
               wrap(PublicationJournal.mark_message_sent)), \
            patch("hop.robust_publisher.PublicationJournal.requeue_message",
                  wrap(PublicationJournal.requeue_message)):
        assert not myLock.was_locked
        callback(err, msg_obj)  # simulate sending failure
        assert not myLock.held, "Lock should be released"
        assert myLock.was_locked, "Lock should have been held"

        # 'get' the message to send again
        s0_2 = j.get_next_message_to_send()
        assert s0_2[0] == s0[0]

        myLock.was_locked = False
        callback(None, msg_obj)  # simulate sending success
        assert not myLock.held, "Lock should be released"
        assert myLock.was_locked, "Lock should have been held"


def test_journal_error_callback():
    # this is just required to not throw
    PublicationJournal.error_callback("terrible error!")


########################################


def makeKafkaException(name="error_name", reason="error reason"):
    err = MagicMock()
    err.name = MagicMock(return_value=name)
    err.reason = MagicMock(return_value=reason)
    err.retriable = MagicMock(return_value=True)
    err.fatal = MagicMock(return_value=False)
    return KafkaException(err)


class FakeProducer:
    def __init__(self, topics, immediate_failure=False, poll_failure=False):
        if len(topics) == 1:
            self.default_topic = topics[0]
        else:
            self.default_topic = None
        self.delivery_callbacks = []
        self.messages_written = []
        # need to coordinate access due to RobustProducer's background thread
        self.lock = threading.Lock()
        # need this to allow ._producer._producer.poll()
        self._producer = MagicMock()
        self._producer._producer = MagicMock()
        if poll_failure:
            logger.debug("poll should fail")
            exc = makeKafkaException()

            def fail_poll(delay):
                time.sleep(delay)
                raise exc

            self._producer._producer.poll = fail_poll
            logger.debug(f" poll is {self._producer._producer.poll}")
        self._immediate_failure = immediate_failure

    def invoke_all_callbacks(self, err, msg):
        with self.lock:
            for callback in self.delivery_callbacks:
                callback(err, msg)
            self.delivery_callbacks.clear()

    def _record_for_topic(self, topic):
        # for now, map any topic to the same producer object
        return hop.io.Producer.TopicRecord(0, 1000000, self._producer)

    def write_raw(self, packed_message, headers, delivery_callback, topic=None, key=None):
        logger.debug(f" FakeProducer.write_raw called with {packed_message}, {headers}")
        if self._immediate_failure:
            self._immediate_failure = False  # stop failing after the first time
            raise makeKafkaException()
        with self.lock:
            self.messages_written.append((topic, packed_message, headers))
            self.delivery_callbacks.append(delivery_callback)

    def flush(self):
        logger.debug(f" FakeProducer.flush called with {len(self.delivery_callbacks)} "
                     "callback(s) in queue")
        # simulate any remaining messages being sent successfully
        msg_no_err = MagicMock()
        msg_no_err.error = MagicMock(return_value=None)
        msg_no_err.latency = MagicMock(return_value=0.1)
        self.invoke_all_callbacks(None, msg_no_err)

    def close(self):
        logger.debug(" FakeProducer.close called")
        self.flush()

    def stop_failing_poll(self):
        self._producer._producer.poll = MagicMock(return_value=None)


# build pointless layers surrounding the hop.io.Producer class
def makeStream(auth=None, *args, **kwargs):
    opener = MagicMock()
    opener.auth = auth or None

    def open(url, mode='r', **o_kwargs):
        username, broker_addresses, topics = parse_kafka_url(url)
        if topics is None:
            topics = []
        return FakeProducer(topics, *args, **kwargs)
    opener.open = open
    return MagicMock(return_value=opener)


def test_rpublisher_empty_journal(tmpdir):
    journal_path = tmpdir.join("journal")
    url = "kafka://example.com/topic"
    fixed_uuid = uuid4()

    with patch("hop.io.Stream", makeStream()) as steam_middleman, \
            patch("hop.io.uuid.uuid4", MagicMock(return_value=fixed_uuid)):
        with RobustProducer(url, journal_path=journal_path, auth=False) as pub:
            pub.write("a message")

            # The publisher will spin in _do_send as long as the state of sent messages is
            # indeterminate. So, spin here in a thread-safe way until the callbacks show up, then
            # invoke them to let it complete.
            while True:
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 1:
                    pub._stream.flush()
                    break
        assert ("topic", *hop.io.Producer.pack("a message", None)) in pub._stream.messages_written


def test_rpublisher_existing_journal(tmpdir):
    journal_path = tmpdir.join("journal")
    url = "kafka://example.com/topic"
    messages = [b"message 0", b"message 1"]

    j = PublicationJournal(journal_path)
    for message in messages:
        j.queue_message(message, "topic")
    del j

    with patch("hop.io.Stream", makeStream()) as steam_middleman:
        with RobustProducer(url, journal_path=journal_path, auth=False) as pub:
            while True:
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 2:
                    pub._stream.flush()
                    break
        # each message previously persisted in the journal should be sent
        sent_messages = [item[1] for item in pub._stream.messages_written]
        for message in messages:
            assert message in sent_messages


def test_rpublisher_with_auth(tmpdir):
    journal_path = tmpdir.join("journal")
    url = "kafka://example.com/topic"
    fixed_uuid = uuid4()
    auth = hop.auth.Auth("user", "password")

    with patch("hop.io.Stream", makeStream(auth=[auth])) as steam_middleman, \
            patch("hop.io.uuid.uuid4", MagicMock(return_value=fixed_uuid)):
        with RobustProducer(url, journal_path=journal_path, auth=True) as pub:
            pub.write("a message")

            # The publisher will spin in _do_send as long as the state of sent messages is
            # indeterminate. So, spin here in a thread-safe way until the callbacks show up, then
            # invoke them to let it complete.
            while True:
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 1:
                    pub._stream.flush()
                    break
        assert ("topic", *hop.io.Producer.pack("a message", None, auth=auth)) \
            in pub._stream.messages_written


def test_rpublisher_automatic_topic(tmpdir):
    journal_path = tmpdir.join("journal")

    # when a single topic is specified in the URL, it should be used as the default by write()
    with patch("hop.io.Stream", makeStream(immediate_failure=True)) as steam_middleman:
        with RobustProducer("kafka://example.com/some_topic",
                            journal_path=journal_path, auth=False) as pub:
            pub.write("a message")

            while True:
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 1:
                    pub._stream.flush()
                    break
            assert pub._stream.messages_written[0][0] == "some_topic"

    # if the URL contains no topic, there is no default and it is an error to call write() without
    # specifying a topic
    with patch("hop.io.Stream", makeStream(immediate_failure=True)) as steam_middleman:
        with RobustProducer("kafka://example.com/", journal_path=journal_path, auth=False) as pub:
            with pytest.raises(Exception):
                pub.write("a message")

    # if the URL contains multiple topics, none of them is treated as a default and it is an error
    # to call write() without specifying a topic
    with patch("hop.io.Stream", makeStream(immediate_failure=True)) as steam_middleman:
        with RobustProducer("kafka://example.com/topicA,topicB",
                            journal_path=journal_path, auth=False) as pub:
            with pytest.raises(Exception):
                pub.write("a message")

            # messages can be written to explicitly specified topics, which need not have been
            # originally listed in the URL
            messages = {"message 1": "topicA", "message 2": "topicB", "message 3": "topicC"}
            for message, topic in messages.items():
                pub.write(message, topic=topic)

            while True:
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 3:
                    pub._stream.flush()
                    break
            for mrecord in pub._stream.messages_written:
                print(mrecord)
                message_data = hop.models.JSONBlob.deserialize(mrecord[1]).content
                assert message_data in messages
                assert mrecord[0] == messages[message_data]


def test_rpublisher_immediate_send_fail(tmpdir):
    journal_path = tmpdir.join("journal")
    url = "kafka://example.com/topic"
    fixed_uuid = uuid4()

    with patch("hop.io.Stream", makeStream(immediate_failure=True)) as steam_middleman, \
            patch("hop.io.uuid.uuid4", MagicMock(return_value=fixed_uuid)):
        with RobustProducer(url, journal_path=journal_path, auth=False) as pub:
            pub.write("a message")

            while True:
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 1:
                    pub._stream.flush()
                    break
        assert ("topic", *hop.io.Producer.pack("a message", None)) in pub._stream.messages_written


def test_rpublisher_poll_fail(tmpdir):
    print("test_rpublisher_poll_fail")
    journal_path = tmpdir.join("journal")
    url = "kafka://example.com/topic"
    fixed_uuid = uuid4()

    with patch("hop.io.Stream", makeStream(poll_failure=True)) as steam_middleman, \
            patch("hop.io.uuid.uuid4", MagicMock(return_value=fixed_uuid)):
        with RobustProducer(url, journal_path=journal_path, auth=False) as pub:
            pub.write("a message")

            while True:
                logger.debug("wating for message to be sent")
                time.sleep(0.01)
                with pub._stream.lock:
                    msg_count = len(pub._stream.messages_written)
                if msg_count == 1:
                    logger.debug("message is visible, waiting to allow poll to proceed")
                    time.sleep(0.01)
                    logger.debug("stopping poll failures")
                    pub._stream.stop_failing_poll()
                    pub._stream.flush()
                    break
            logger.debug("with body done")
        assert ("topic", *hop.io.Producer.pack("a message", None)) in pub._stream.messages_written



from functools import partial
import logging
import pytest
import requests
import scramp
from unittest.mock import patch, MagicMock

from conftest import PhonyConnection, mock_pool_manager

from hop.auth import Auth
from hop.http_scram import SCRAMAuth

logger = logging.getLogger("hop")

# Replicate the example SCRAM exchange from RFC 5802
test_mechanism = "SCRAM-SHA-1"
test_auth = Auth("user", "pencil", method=test_mechanism)
test_nonce = "fyko+d2lbbFgONRv9qkxdawL"
test_sid = "AAAABBBBCCCCDDDD"
test_client_first = "SCRAM-SHA-1 data=biwsbj11c2VyLHI9ZnlrbytkMmxiYkZnT05Sdjlxa3hkYXdM"
test_server_first_data = "cj1meWtvK2QybGJiRmdPTlJ2OXFreGRhd0wzcmZjTkhZSlkxWlZ2V1ZzN" \
                         "2oscz1RU1hDUitRNnNlazhiZjkyLGk9NDA5Ng=="
test_server_first = f"SCRAM-SHA-1 sid={test_sid},data={test_server_first_data}"
test_client_final_data = "Yz1iaXdzLHI9ZnlrbytkMmxiYkZnT05Sdjlxa3hkYXdMM3JmY05IWUpZ" \
                         "MVpWdldWczdqLHA9djBYOHYzQnoyVDBDSkdiSlF5RjBYK0hJNFRzPQ=="
test_client_final = f"SCRAM-SHA-1 sid={test_sid},data={test_client_final_data}"
test_server_final = f"sid={test_sid},data=dj1ybUY5cHFWOFM3c3VBb1pXamE0ZEpSa0ZzS1E9"


def test_SCRAMAuth_generate_client_first():
    with patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth)
        cf = a._generate_client_first()
        assert cf == test_client_first


def request_copy(self):
    new_request = MagicMock()
    new_request.headers = self.headers
    new_request.url = self.url
    return new_request


def test_SCRAMAuth_handle_first():
    with patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        # matching mechanism
        a = SCRAMAuth(test_auth)
        test_resp = MagicMock()
        test_resp.headers = {"www-authenticate": test_mechanism}
        test_resp.request.url = "http://example.com"
        test_resp.request.headers = {}
        test_resp.request.copy = partial(request_copy, test_resp.request)
        result = a._handle_first(test_resp)
        assert "Authorization" in result.request.headers
        assert result.request.headers["Authorization"] == test_client_first

        # non-matching mechanism
        test_resp = MagicMock()
        test_resp.headers = {"www-authenticate": "UNEXPECTED-MECHANISM"}
        test_resp.request.url = "http://example.com"
        test_resp.request.headers = {}
        test_resp.request.copy = partial(request_copy, test_resp.request)
        result = a._handle_first(test_resp)
        assert "Authorization" not in result.request.headers


def test_SCRAMAuth_handle_final():
    with patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        # matching mechanism
        a = SCRAMAuth(test_auth)
        test_resp = MagicMock()
        test_resp.headers = {"www-authenticate": test_server_first}
        test_resp.request.url = "http://example.com"
        test_resp.request.headers = {}
        test_resp.request.copy = partial(request_copy, test_resp.request)
        a._generate_client_first()
        result = a._handle_final(test_resp)
        assert "Authorization" in result.request.headers
        assert result.request.headers["Authorization"] == test_client_final

        # non-matching mechanism
        test_server_first_wrong_mech = f"UNEXPECTED-MECHANISM sid={test_sid}," \
            f"data={test_server_first_data}"
        a = SCRAMAuth(test_auth)
        test_resp = MagicMock()
        test_resp.headers = {"www-authenticate": test_server_first_wrong_mech}
        test_resp.request.url = "http://example.com"
        test_resp.request.headers = {}
        test_resp.request.copy = partial(request_copy, test_resp.request)
        a._generate_client_first()
        result = a._handle_final(test_resp)
        assert "Authorization" not in result.request.headers

        # missing sid
        test_server_first_no_sid = f"SCRAM-SHA-1 data={test_server_first_data}"
        a = SCRAMAuth(test_auth)
        test_resp = MagicMock()
        test_resp.headers = {"www-authenticate": test_server_first_no_sid}
        test_resp.request.url = "http://example.com"
        test_resp.request.headers = {}
        test_resp.request.copy = partial(request_copy, test_resp.request)
        a._generate_client_first()
        with pytest.raises(RuntimeError) as err:
            result = a._handle_final(test_resp)
            assert "Authorization" not in result.request.headers

        # missing data
        test_server_first_no_data = f"SCRAM-SHA-1 sid={test_sid}"
        a = SCRAMAuth(test_auth)
        test_resp = MagicMock()
        test_resp.headers = {"www-authenticate": test_server_first_no_data}
        test_resp.request.url = "http://example.com"
        test_resp.request.headers = {}
        test_resp.request.copy = partial(request_copy, test_resp.request)
        a._generate_client_first()
        with pytest.raises(RuntimeError) as err:
            result = a._handle_final(test_resp)
            assert "Authorization" not in result.request.headers


def test_SCRAMAuth_check_server_final():
    with patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth)
        sfirst = MagicMock()
        sfirst.headers = {"www-authenticate": test_server_first}
        sfirst.request.url = "http://example.com"
        sfirst.request.headers = {}
        sfirst.request.copy = partial(request_copy, sfirst.request)
        a._generate_client_first()
        a._handle_final(sfirst)
        sfinal = MagicMock()
        sfinal.headers = {"authentication-info": test_server_final}
        sfinal.request.url = "http://example.com"
        sfinal.request.headers = {}
        sfinal.request.copy = partial(request_copy, sfinal.request)
        a._check_server_final(sfinal)

        # missing sid
        test_server_final_no_sid = "data=dj1ybUY5cHFWOFM3c3VBb1pXamE0ZEpSa0ZzS1E9"
        a = SCRAMAuth(test_auth)
        sfirst = MagicMock()
        sfirst.headers = {"www-authenticate": test_server_first}
        sfirst.request.url = "http://example.com"
        sfirst.request.headers = {}
        sfirst.request.copy = partial(request_copy, sfirst.request)
        a._generate_client_first()
        a._handle_final(sfirst)
        sfinal = MagicMock()
        sfinal.headers = {"authentication-info": test_server_final_no_sid}
        sfinal.request.url = "http://example.com"
        sfinal.request.headers = {}
        sfinal.request.copy = partial(request_copy, sfinal.request)
        with pytest.raises(RuntimeError):
            a._check_server_final(sfinal)

        # wrong sid
        test_server_final_wrong_sid = "sid=XYZ,data=dj1ybUY5cHFWOFM3c3VBb1pXamE0ZEpSa0ZzS1E9"
        a = SCRAMAuth(test_auth)
        sfirst = MagicMock()
        sfirst.headers = {"www-authenticate": test_server_first}
        sfirst.request.url = "http://example.com"
        sfirst.request.headers = {}
        sfirst.request.copy = partial(request_copy, sfirst.request)
        a._generate_client_first()
        a._handle_final(sfirst)
        sfinal = MagicMock()
        sfinal.headers = {"authentication-info": test_server_final_wrong_sid}
        sfinal.request.url = "http://example.com"
        sfinal.request.headers = {}
        sfinal.request.copy = partial(request_copy, sfinal.request)
        with pytest.raises(RuntimeError):
            a._check_server_final(sfinal)

        # missing data
        test_server_final_no_data = f"sid={test_sid}"
        a = SCRAMAuth(test_auth)
        sfirst = MagicMock()
        sfirst.headers = {"www-authenticate": test_server_first}
        sfirst.request.url = "http://example.com"
        sfirst.request.headers = {}
        sfirst.request.copy = partial(request_copy, sfirst.request)
        a._generate_client_first()
        a._handle_final(sfirst)
        sfinal = MagicMock()
        sfinal.headers = {"authentication-info": test_server_final_no_data}
        sfinal.request.url = "http://example.com"
        sfinal.request.headers = {}
        sfinal.request.copy = partial(request_copy, sfinal.request)
        with pytest.raises(RuntimeError):
            a._check_server_final(sfinal)

        # wrong data
        test_server_final_wrong_data = f"sid={test_sid},data=dj1iYWRkYXRh"
        a = SCRAMAuth(test_auth)
        sfirst = MagicMock()
        sfirst.headers = {"www-authenticate": test_server_first}
        sfirst.request.url = "http://example.com"
        sfirst.request.headers = {}
        sfirst.request.copy = partial(request_copy, sfirst.request)
        a._generate_client_first()
        a._handle_final(sfirst)
        sfinal = MagicMock()
        sfinal.headers = {"authentication-info": test_server_final_wrong_data}
        sfinal.request.url = "http://example.com"
        sfinal.request.headers = {}
        sfinal.request.copy = partial(request_copy, sfinal.request)
        with pytest.raises(scramp.core.ScramException):
            a._check_server_final(sfinal)


def test_SCRAMAuth_integration():
    conn = PhonyConnection([MagicMock(status=401,
                                      headers={"WWW-Authenticate":
                                               f'{test_mechanism} realm="realm@example.com"'}
                                      ),
                            MagicMock(status=401,
                                      headers={"WWW-Authenticate": test_server_first}
                                      ),
                            MagicMock(status=200,
                                      headers={"Authentication-Info": test_server_final}
                                      )
                            ])
    PM = mock_pool_manager(conn)
    with patch("requests.adapters.PoolManager", PM), \
            patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth)
        resp = requests.get("http://example.com", auth=a)
        print(resp)
        assert conn.counter == 3
        assert "Authorization" not in conn.requests[0]["headers"]
        assert "Authorization" in conn.requests[1]["headers"]
        assert conn.requests[1]["headers"]["Authorization"] == test_client_first
        assert "Authorization" in conn.requests[2]["headers"]
        assert conn.requests[2]["headers"]["Authorization"] == test_client_final


def test_SCRAMAuth_integration_shortcut():
    conn = PhonyConnection([MagicMock(status=401,
                                      headers={"WWW-Authenticate": test_server_first}
                                      ),
                            MagicMock(status=200,
                                      headers={"Authentication-Info": test_server_final}
                                      )
                            ])
    PM = mock_pool_manager(conn)
    with patch("requests.adapters.PoolManager", PM), \
            patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth, shortcut=True)
        body = "A" * 1024
        resp = requests.post("http://example.com", auth=a, data=body)
        assert conn.counter == 2
        assert "Authorization" in conn.requests[0]["headers"]
        assert conn.requests[0]["headers"]["Authorization"] == test_client_first
        assert conn.requests[0]["body"] is None, "No body should be sent during auth handshake"
        assert "Authorization" in conn.requests[1]["headers"]
        assert conn.requests[1]["headers"]["Authorization"] == test_client_final
        assert conn.requests[1]["body"] == body, "After handshake completes, body should be sent"


def test_SCRAMAuth_integration_auth_unnecessary():
    conn = PhonyConnection([MagicMock(status=200,
                                      headers={},
                                      ),
                            ])
    PM = mock_pool_manager(conn)
    with patch("requests.adapters.PoolManager", PM), \
            patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth)
        resp = requests.get("http://example.com", auth=a)
        assert conn.counter == 1
        assert "Authorization" not in conn.requests[0]["headers"]


def test_SCRAMAuth_integration_bad_server_final():
    conn = PhonyConnection([MagicMock(status=401,
                                      headers={"WWW-Authenticate":
                                               f'{test_mechanism} realm="realm@example.com"'}
                                      ),
                            MagicMock(status=401,
                                      headers={"WWW-Authenticate": test_server_first}
                                      ),
                            MagicMock(status=200,
                                      headers={"Authentication-Info":
                                               f"sid={test_sid},data=dj1CYWRTZXJ2ZXJTaWduYXR1cmU="}
                                      )
                            ])
    PM = mock_pool_manager(conn)
    with patch("requests.adapters.PoolManager", PM), \
            patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth)
        with pytest.raises(scramp.ScramException):
            resp = requests.get("http://example.com", auth=a)


def test_SCRAMAuth_integration_server_fail():
    conn = PhonyConnection([MagicMock(status=401,
                                      headers={"WWW-Authenticate":
                                               f'{test_mechanism} realm="realm@example.com"'}
                                      ),
                            MagicMock(status=500)
                            ])
    PM = mock_pool_manager(conn)
    with patch("requests.adapters.PoolManager", PM), \
            patch("secrets.token_urlsafe", MagicMock(return_value=test_nonce)):
        a = SCRAMAuth(test_auth)
        resp = requests.get("http://example.com", auth=a)
        assert resp.status_code == 500
        assert conn.counter == 2
        assert "Authorization" not in conn.requests[0]["headers"]
        assert "Authorization" in conn.requests[1]["headers"]
        assert conn.requests[1]["headers"]["Authorization"] == test_client_first



import argparse
import logging
import sys
import traceback

from . import __version__
from . import configure
from . import auth
from . import publish
from . import subscribe
from . import version
from . import list_topics
from .utils import cli as cli_utils


logger = logging.getLogger("hop")


def set_up_cli():
    """Set up CLI commands for hop entry point.

    Returns:
        An ArgumentParser instance with hop-based commands and relevant
        arguments added for all commands.

    """
    parser = argparse.ArgumentParser(prog="hop", formatter_class=cli_utils.SubcommandHelpFormatter)
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s version {__version__}",
    )

    parser.add_argument(
        "-d", "--debug", action="store_true",
        help="Display more verbose errors with internal details",
    )

    # set up subparser
    subparser = parser.add_subparsers(title="commands", metavar="<command>", dest="cmd")
    subparser.required = True

    # register commands
    p = cli_utils.append_subparser(
        subparser,
        "configure",
        configure._main,
        formatter_class=cli_utils.SubcommandHelpFormatter
    )
    configure._add_parser_args(p)

    p = cli_utils.append_subparser(subparser, "auth", auth._main)
    auth._add_parser_args(p)

    p = cli_utils.append_subparser(subparser, "publish", publish._main)
    publish._add_parser_args(p)

    p = cli_utils.append_subparser(subparser, "subscribe", subscribe._main)
    subscribe._add_parser_args(p)

    p = cli_utils.append_subparser(subparser, "version", version._main)

    p = cli_utils.append_subparser(subparser, "list-topics", list_topics._main)
    list_topics._add_parser_args(p)

    return parser


def check_auth_data(prog_name):
    """Advise the user on getting/configuring credential data
    """

    advice = """
No valid credential data found
You can get a credential from https://my.hop.scimma.org
To load your credential, run `hop auth add`
"""

    try:
        auth.load_auth()
    except FileNotFoundError:
        print(advice)
    except Exception as ex:
        print(prog_name + ":", ex, file=sys.stderr)
        print(advice)


def main():
    parser = set_up_cli()

    try:
        args = parser.parse_args()
    except SystemExit:
        check_auth_data(parser.prog)
        raise
    try:
        args.func(args)
    except KeyboardInterrupt:
        logger.info("received keyboard interrupt, closing")
    except Exception as ex:
        if args.debug:
            traceback.print_exc(file=sys.stderr)
        else:
            print(parser.prog + ":", ex, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import pluggy


# hooks for registering plugins
register = pluggy.HookimplMarker("hop")
specification = pluggy.HookspecMarker("hop")


@specification
def get_models():
    """
    This plugin spec is used to return message models in the form:
        {"type": Model}

    where the type refers to a specific message model.

    """



from unittest.mock import patch, mock_open
from io import BytesIO
import json
from pathlib import Path
import pytest

import fastavro

from hop import models
from conftest import message_parameters_dict_data


def test_MessageModel():
    # derived classes must implement their own load() logic, and the base implementation has no
    # useful functionality
    with pytest.raises(NotImplementedError):
        models.MessageModel.load(None)


def test_voevent(voevent_fileobj):
    voevent = models.VOEvent.load(voevent_fileobj)

    # check a few attributes
    assert voevent.ivorn == "ivo://gwnet/LVC#S200302c-1-Preliminary"
    assert voevent.role == "observation"
    assert voevent.version == "2.0"

    assert voevent.Who["Date"] == "2020-03-02T02:00:09"
    assert voevent.Description == "Report of a candidate gravitational wave event"
    assert voevent.WhereWhen["ObsDataLocation"]["ObservatoryLocation"]["id"] == "LIGO Virgo"

    # verify wrapper format is correct
    assert voevent.serialize()["format"] == "voevent"

    # conversion to a string should produce valid JSON which should parse to match the original
    # event data
    as_string = str(voevent)
    reparsed = json.loads(as_string)
    assert reparsed["ivorn"] == voevent.ivorn
    assert reparsed["role"] == voevent.role
    assert reparsed["version"] == voevent.version
    assert reparsed["Who"]["Date"] == voevent.Who["Date"]
    assert reparsed["Description"] == voevent.Description
    assert reparsed["WhereWhen"]["ObsDataLocation"]["ObservatoryLocation"]["id"] \
        == voevent.WhereWhen["ObsDataLocation"]["ObservatoryLocation"]["id"]


def test_voevent_deserialization():
    mpdd = message_parameters_dict_data

    v2 = models.VOEvent.deserialize(mpdd["voevent"]["model_text"])
    assert isinstance(v2, models.VOEvent)

    v1 = models.VOEvent.deserialize(mpdd["voevent-encoded-as-json"]["model_text"])
    assert isinstance(v1, models.VOEvent)

    with pytest.raises(Exception) as ex:
        vx = models.VOEvent.deserialize(b"\tNot XML or JSON\t")


def test_voevent_dict_factory():
    d1 = {"foo": 1, "bar": 2, "baz": 3}
    o1 = models.VOEvent.dict_factory(d1.items())
    assert o1 == d1

    d2 = {"foo": 1, "_bar": 2, "baz": 3, "_quux": 4}
    o2 = models.VOEvent.dict_factory(d2.items())
    assert len(o2) == 2
    assert "foo" in o2
    assert "_bar" not in o2
    assert "baz" in o2
    assert "_quux" not in o2


VOEvent_attribute_names = {"cite", "coord_system_id", "dataType", "expires", "id", "importance",
                           "ivorn", "meaning", "mimetype", "name", "probability", "relation",
                           "role", "type", "ucd", "unit", "uri", "utype", "value", "version"}


VOEvent_element_names = {"AstroCoordSystem", "AstroCoords", "Author", "AuthorIVORN", "C1", "C2",
                         "C3", "Citations", "Concept", "Data", "Date", "Description", "Error",
                         "Error2Radius", "EventIVORN", "Field", "Group", "How", "ISOTime",
                         "Inference", "Name", "Name1", "Name2", "Name3", "ObsDataLocation",
                         "ObservationLocation", "ObservatoryLocation", "Param", "Position2D",
                         "Position3D", "Reference", "TD", "TR", "Table", "Time", "TimeInstant",
                         "TimeOffset", "TimeScale", "VOEvent", "Value", "Value2", "Value3",
                         "What", "WhereWhen", "Who", "Why", "contactEmail", "contactName",
                         "contactPhone", "contributor", "logoURL", "shortName", "title"}


def test_voevent_label_attributes_list():
    ta = [{n: 0} for n in VOEvent_attribute_names]
    models.VOEvent.label_attributes_list(ta)
    for entry in ta:
        for key in entry.keys():
            assert key.startswith('@')
            assert key[1:] in VOEvent_attribute_names

    nta = [[{n: 0} for n in VOEvent_attribute_names]]
    models.VOEvent.label_attributes_list(nta)
    assert len(nta) == 1
    for entry in nta[0]:
        for key in entry.keys():
            assert key.startswith('@')
            assert key[1:] in VOEvent_attribute_names

    te = [{n: 0} for n in VOEvent_element_names]
    models.VOEvent.label_attributes_list(te)
    for entry in te:
        for key in entry.keys():
            assert not key.startswith('@')
            assert key in VOEvent_element_names

    nte = [[{n: 0} for n in VOEvent_element_names]]
    models.VOEvent.label_attributes_list(nte)
    assert len(nte) == 1
    for entry in nte[0]:
        for key in entry.keys():
            assert not key.startswith('@')
            assert key in VOEvent_element_names


def test_voevent_label_attributes_dict():
    ta = {n: 0 for n in VOEvent_attribute_names}
    models.VOEvent.label_attributes_dict(ta)
    for key in ta.keys():
        assert key.startswith('@')
        assert key[1:] in VOEvent_attribute_names

    te = {n: 0 for n in VOEvent_element_names}
    models.VOEvent.label_attributes_dict(te)
    for key in te.keys():
        assert not key.startswith('@')
        assert key in VOEvent_element_names


def test_voevent_ensure_bytes():
    b = b"abcdef\x00\x01ghij"
    assert models.VOEvent.ensure_bytes(b) == b

    s = "foo bar baz"
    assert models.VOEvent.ensure_bytes(s) == s.encode("utf-8")

    f = BytesIO(b)
    assert models.VOEvent.ensure_bytes(f) == b

    assert models.VOEvent.ensure_bytes(c.to_bytes(1, byteorder="big") for c in b) == b


def test_gcn_text_notice(gcn_text_notice_fileobj, gcn_text_notice_data):
    text_notice = models.GCNTextNotice.load(gcn_text_notice_fileobj)

    # the object should retain the raw data unchanged for lossless round-tripping
    assert text_notice.raw == gcn_text_notice_data

    # check the data was parsed sensibly
    expected_fields = ["title", "notice_date", "notice_type", "stream", "run_num", "event_num",
                       "src_ra", "src_dec", "src_error", "src_error50", "discovery_date",
                       "discovery_time", "revision", "energy", "signalness", "far", "sun_postn",
                       "sun_dist", "moon_postn", "moon_dist", "gal_coords", "ecl_coords",
                       "comments"]
    for field in expected_fields:
        assert field in text_notice.fields

    expected_title = "GCN/AMON NOTICE"
    assert text_notice.fields["title"] == expected_title
    expected_energy = "1.2126e+02 [TeV]"
    assert text_notice.fields["energy"] == expected_energy
    # multi-line values should be accumulated
    expected_src_ra = \
        "103.7861d {+06h 55m 09s} (J2000),\n" \
        "104.1106d {+06h 56m 27s} (current),\n" \
        "103.1176d {+06h 52m 28s} (1950)"
    assert text_notice.fields["src_ra"] == expected_src_ra
    # repeated keys should have their values accumulated
    expected_comments = \
        "IceCube Bronze event.\n" \
        "The position error is statistical only, there is no systematic added."
    assert text_notice.fields["comments"] == expected_comments


def test_gcn_circular(circular_text, circular_msg, circular_data_raw):
    with patch("builtins.open", mock_open(read_data=circular_text)):
        gcn_file = "example.gcn3"
        with open(gcn_file, "r") as f:
            gcn = models.GCNCircular.load(f)

        # verify parsed GCN structure is correct
        assert gcn.header["title"] == circular_data_raw["header"]["title"]
        assert gcn.header["number"] == circular_data_raw["header"]["number"]
        assert gcn.header["subject"] == circular_data_raw["header"]["subject"]
        assert gcn.header["date"] == circular_data_raw["header"]["date"]
        assert gcn.header["from"] == circular_data_raw["header"]["from"]

        assert gcn.body == circular_data_raw["body"]

        # verify wrapper format is correct
        assert gcn.serialize()["format"] == "circular"

        # conversion to a string should produce email formatted data which should parse to match the
        # original event data
        as_string = str(gcn)
        reparsed = models.GCNCircular.load(as_string)
        assert reparsed.header == gcn.header
        assert reparsed.body == gcn.body


def test_blob(blob_text, blob_msg):
    with patch("builtins.open", mock_open(read_data=blob_text)):
        blob_file = "example_blob.txt"
        blob = models.Blob.load_file(blob_file)

        # verify blob text is correct
        assert blob.content == blob_msg["content"]

        # verify wrapper format is correct
        assert blob.serialize()["format"] == "blob"

        assert bytes(blob) == blob_msg["content"]
        assert models.Blob.load(bytes(blob)) == blob


def test_avro(avro_data_raw, avro_data, avro_msg):
    with patch("builtins.open", mock_open(read_data=avro_data)):
        avro = models.AvroBlob.load(avro_data)

    # verify data is correct after deserializing
    assert avro.content == avro_data_raw

    # verify wrapper format is correct
    assert avro.serialize()["format"] == "avro"

    # verify that serializing again produces the same result
    serialized = avro.serialize()["content"]
    assert models.AvroBlob.load(serialized) == avro

    # test single_record = False codepath
    with patch("builtins.open", mock_open(read_data=avro_data)):
        avro2 = models.AvroBlob.load(avro_data, single_record=False)

    # verify data is correct after deserializing
    assert avro2.content == [avro_data_raw]

    # verify wrapper format is correct
    assert avro2.serialize()["format"] == "avro"

    # verify that serializing again produces the same result
    serialized2 = avro2.serialize()["content"]
    assert models.AvroBlob.load(serialized2, single_record=False) == avro2

    # test that serializing the same object tree without a schema produces equivalent results
    avro3 = models.AvroBlob(avro_data_raw)
    serialized3 = avro3.serialize()["content"]

    # serialized2 may not be the same as avro_data because the schemas can differ
    assert models.AvroBlob.load(serialized3) == avro3


def test_avro_invalid_data_type_serialization():
    # data must be a sequence of records if single_record = False
    with pytest.raises(TypeError) as te:
        models.AvroBlob({1: "abc", False: "def"}, single_record=False).serialize()
    assert "AvroBlob requires content to be a sequence of records when " \
           "single_record = False" in str(te.value)

    # dictionary keys can only be strings
    with pytest.raises(ValueError) as ve:
        models.AvroBlob([{1: "abc", False: "def"}]).serialize()
    assert "Dictionaries with non-string keys cannot be represented as Avro" in str(ve.value)

    # arbitrary, user-defined types are not supported in Avro
    with pytest.raises(ValueError) as ve:
        models.AvroBlob([models.Blob(b"somedata")]).serialize()
    assert "Unable to assign an Avro type to value of type <class 'hop.models.Blob'>" \
        in str(ve.value)


def test_avro_invalid_data_type_deserialization():
    # only bytes objects are accepted, not strings
    with pytest.raises(TypeError):
        models.AvroBlob.load("some string")


def test_avro_operators(avro_data, avro_data_raw):
    avro_with_schema = models.AvroBlob.load(avro_data)
    avro_no_schema = models.AvroBlob(avro_data_raw)

    # equality should not consider schemas
    assert avro_with_schema == avro_no_schema
    # not equal to objects of any other type
    assert avro_with_schema != avro_data_raw
    # hashing is not implemented because python fails to implement it for dict
    with pytest.raises(NotImplementedError):
        hash(avro_with_schema)


def test_avro_multi_record_deserialization():
    original = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    outio = BytesIO()
    schema = {'type': 'array', 'items': 'long'}
    fastavro.writer(outio, schema, original)
    raw = outio.getvalue()

    # default behavior should be to require a single record
    with pytest.raises(TypeError) as te:
        loaded = models.AvroBlob.load(raw)
    assert "AvroBlob requires content to be a single record" in str(te.value)

    # allowing multiple records should load as such
    loaded = models.AvroBlob.load(raw, single_record=False)
    assert loaded.content == original


def test_externalmessage():
    inputs = [
        '{"url": "http://example.com"}',
        b'{"url": "http://example.com"}',
    ]
    for input in inputs:
        for read in [True, False]:
            if read:
                with patch("builtins.open", mock_open(read_data=input)):
                    m = models.ExternalMessage.load(open("notarealfile"))
            else:
                m = models.ExternalMessage.load(input)
            assert m.url == "http://example.com"

    with pytest.raises(json.JSONDecodeError):
        models.ExternalMessage.load("\tnot valid JSON\t")


@pytest.mark.parametrize("model", [
    models.Blob,
    models.JSONBlob,
    models.AvroBlob,
    models.GCNTextNotice,
    models.GCNCircular,
    models.VOEvent,
])
def test_model_roundtrip(model, tmpdir):
    model_name = models.format_name(model)
    shared_datadir = Path("tests/data")
    orig_path = shared_datadir / "test_data" / message_parameters_dict_data[model_name]["test_file"]
    instance1 = model.load_file(orig_path)
    test_path = tmpdir.join("rewritten.dat")
    with open(test_path, "wb") as out_file:
        out_file.write(bytes(instance1))
    instance2 = model.load_file(test_path)
    assert instance1 == instance2
